// scripts/main.ts
import {
  system,
  world,
  Entity,
  EquipmentSlot
} from "@minecraft/server";

// node_modules/@minecraft/math/lib/general/clamp.js
function clampNumber(val, min, max) {
  return Math.min(Math.max(val, min), max);
}

// node_modules/@minecraft/math/lib/vector3/coreHelpers.js
var Vector3Utils = class _Vector3Utils {
  /**
   * equals
   *
   * Check the equality of two vectors
   */
  static equals(v1, v2) {
    return v1.x === v2.x && v1.y === v2.y && v1.z === v2.z;
  }
  /**
   * add
   *
   * Add two vectors to produce a new vector
   */
  static add(v1, v2) {
    return { x: v1.x + v2.x, y: v1.y + v2.y, z: v1.z + v2.z };
  }
  /**
   * subtract
   *
   * Subtract two vectors to produce a new vector (v1-v2)
   */
  static subtract(v1, v2) {
    return { x: v1.x - v2.x, y: v1.y - v2.y, z: v1.z - v2.z };
  }
  /** scale
   *
   * Multiple all entries in a vector by a single scalar value producing a new vector
   */
  static scale(v1, scale) {
    return { x: v1.x * scale, y: v1.y * scale, z: v1.z * scale };
  }
  /**
   * dot
   *
   * Calculate the dot product of two vectors
   */
  static dot(a, b) {
    return a.x * b.x + a.y * b.y + a.z * b.z;
  }
  /**
   * cross
   *
   * Calculate the cross product of two vectors. Returns a new vector.
   */
  static cross(a, b) {
    return {
      x: a.y * b.z - a.z * b.y,
      y: a.z * b.x - a.x * b.z,
      z: a.x * b.y - a.y * b.x
    };
  }
  /**
   * magnitude
   *
   * The magnitude of a vector
   */
  static magnitude(v) {
    return Math.sqrt(v.x ** 2 + v.y ** 2 + v.z ** 2);
  }
  /**
   * distance
   *
   * Calculate the distance between two vectors
   */
  static distance(a, b) {
    return _Vector3Utils.magnitude(_Vector3Utils.subtract(a, b));
  }
  /**
   * normalize
   *
   * Takes a vector 3 and normalizes it to a unit vector
   */
  static normalize(v) {
    const mag = _Vector3Utils.magnitude(v);
    return { x: v.x / mag, y: v.y / mag, z: v.z / mag };
  }
  /**
   * floor
   *
   * Floor the components of a vector to produce a new vector
   */
  static floor(v) {
    return { x: Math.floor(v.x), y: Math.floor(v.y), z: Math.floor(v.z) };
  }
  /**
   * toString
   *
   * Create a string representation of a vector3
   */
  static toString(v, options) {
    const decimals = options?.decimals ?? 2;
    const str = [v.x.toFixed(decimals), v.y.toFixed(decimals), v.z.toFixed(decimals)];
    return str.join(options?.delimiter ?? ", ");
  }
  /**
   * clamp
   *
   * Clamps the components of a vector to limits to produce a new vector
   */
  static clamp(v, limits) {
    return {
      x: clampNumber(v.x, limits?.min?.x ?? Number.MIN_SAFE_INTEGER, limits?.max?.x ?? Number.MAX_SAFE_INTEGER),
      y: clampNumber(v.y, limits?.min?.y ?? Number.MIN_SAFE_INTEGER, limits?.max?.y ?? Number.MAX_SAFE_INTEGER),
      z: clampNumber(v.z, limits?.min?.z ?? Number.MIN_SAFE_INTEGER, limits?.max?.z ?? Number.MAX_SAFE_INTEGER)
    };
  }
  /**
   * lerp
   *
   * Constructs a new vector using linear interpolation on each component from two vectors.
   */
  static lerp(a, b, t) {
    return {
      x: a.x + (b.x - a.x) * t,
      y: a.y + (b.y - a.y) * t,
      z: a.z + (b.z - a.z) * t
    };
  }
  /**
   * slerp
   *
   * Constructs a new vector using spherical linear interpolation on each component from two vectors.
   */
  static slerp(a, b, t) {
    const theta = Math.acos(_Vector3Utils.dot(a, b));
    const sinTheta = Math.sin(theta);
    const ta = Math.sin((1 - t) * theta) / sinTheta;
    const tb = Math.sin(t * theta) / sinTheta;
    return _Vector3Utils.add(_Vector3Utils.scale(a, ta), _Vector3Utils.scale(b, tb));
  }
};

// node_modules/@minecraft/math/lib/vector3/vectorWrapper.js
var Vector3Builder = class {
  constructor(first, y, z) {
    if (typeof first === "object") {
      this.x = first.x;
      this.y = first.y;
      this.z = first.z;
    } else {
      this.x = first;
      this.y = y ?? 0;
      this.z = z ?? 0;
    }
  }
  /**
   * Assigns the values of the passed in vector to this vector. Returns itself.
   */
  assign(vec) {
    this.x = vec.x;
    this.y = vec.y;
    this.z = vec.z;
    return this;
  }
  /**
   * equals
   *
   * Check the equality of two vectors
   */
  equals(v) {
    return Vector3Utils.equals(this, v);
  }
  /**
   * add
   *
   * Adds the vector v to this, returning itself.
   */
  add(v) {
    return this.assign(Vector3Utils.add(this, v));
  }
  /**
   * subtract
   *
   * Subtracts the vector v from this, returning itself.
   */
  subtract(v) {
    return this.assign(Vector3Utils.subtract(this, v));
  }
  /** scale
   *
   * Scales this by the passed in value, returning itself.
   */
  scale(val) {
    return this.assign(Vector3Utils.scale(this, val));
  }
  /**
   * dot
   *
   * Computes the dot product of this and the passed in vector.
   */
  dot(vec) {
    return Vector3Utils.dot(this, vec);
  }
  /**
   * cross
   *
   * Computes the cross product of this and the passed in vector, returning itself.
   */
  cross(vec) {
    return this.assign(Vector3Utils.cross(this, vec));
  }
  /**
   * magnitude
   *
   * The magnitude of the vector
   */
  magnitude() {
    return Vector3Utils.magnitude(this);
  }
  /**
   * distance
   *
   * Calculate the distance between two vectors
   */
  distance(vec) {
    return Vector3Utils.distance(this, vec);
  }
  /**
   * normalize
   *
   * Normalizes this vector, returning itself.
   */
  normalize() {
    return this.assign(Vector3Utils.normalize(this));
  }
  /**
   * floor
   *
   * Floor the components of a vector to produce a new vector
   */
  floor() {
    return this.assign(Vector3Utils.floor(this));
  }
  /**
   * toString
   *
   * Create a string representation of a vector
   */
  toString(options) {
    return Vector3Utils.toString(this, options);
  }
  /**
   * clamp
   *
   * Clamps the components of a vector to limits to produce a new vector
   */
  clamp(limits) {
    return this.assign(Vector3Utils.clamp(this, limits));
  }
  /**
   * lerp
   *
   * Constructs a new vector using linear interpolation on each component from two vectors.
   */
  lerp(vec, t) {
    return this.assign(Vector3Utils.lerp(this, vec, t));
  }
  /**
   * slerp
   *
   * Constructs a new vector using spherical linear interpolation on each component from two vectors.
   */
  slerp(vec, t) {
    return this.assign(Vector3Utils.slerp(this, vec, t));
  }
};

// scripts/MathUtils.ts
var MathUtils = class _MathUtils {
  //---------------------------------------
  //  applyDirectionalOffsetToCoords
  //  - https://discord.com/channels/523663022053392405/1131984839151714354
  //------------------------------------------------------------------------------
  static applyDirectionalOffsetToCoords(baseCoords, offsetCoordsStr, viewDir) {
    let coords = offsetCoordsStr.split(" ");
    let relativeCoordVector = { x: 0, y: 0, z: 0 };
    let localCoords = { x: 0, y: 0, z: 0 };
    let absoluteCoords = baseCoords;
    coords.forEach((coord, i) => {
      if (coord.substring(0, 1) == "^") {
        let curCoord = 0;
        if (coord.length > 1)
          curCoord = parseFloat(coord.substring(1));
        if (i == 0)
          localCoords.x = curCoord;
        else if (i == 1)
          localCoords.y = curCoord;
        else
          localCoords.z = curCoord;
      } else if (coord.substring(0, 1) == "~") {
        let curCoord = 0;
        if (coord.length > 1)
          curCoord = parseFloat(coord.substring(1));
        if (i == 0)
          relativeCoordVector.x = curCoord;
        else if (i == 1)
          relativeCoordVector.y = curCoord;
        else
          relativeCoordVector.z = curCoord;
      } else {
        let curCoord = parseFloat(coord);
        if (i == 0)
          absoluteCoords.x = curCoord;
        else if (i == 1)
          absoluteCoords.y = curCoord;
        else
          absoluteCoords.z = curCoord;
      }
    });
    absoluteCoords = Vector3Utils.add(absoluteCoords, relativeCoordVector);
    let zVec = new Vector3Builder(viewDir);
    let xVec = _MathUtils.normalizeVector(new Vector3Builder(zVec.z, 0, -zVec.x));
    let yVec = _MathUtils.normalizeVector(new Vector3Builder(Vector3Utils.cross(zVec, xVec)));
    let xLocationOffset = xVec.scale(localCoords.x);
    let yLocationOffset = yVec.scale(localCoords.y);
    let zLocationOffset = zVec.scale(localCoords.z);
    let locationOffset = Vector3Utils.add(Vector3Utils.add(xLocationOffset, yLocationOffset), zLocationOffset);
    absoluteCoords = Vector3Utils.add(absoluteCoords, locationOffset);
    return new Vector3Builder(absoluteCoords);
  }
  //---------------------------------------
  //  normalizeVector
  //  reduce vector to length 1 (unit vector), and adjust x/y/z to match
  //------------------------------------------------------------------------------
  static normalizeVector(v) {
    let mag = Vector3Utils.magnitude(v);
    let x = v.x / mag;
    let y = v.y / mag;
    let z = v.z / mag;
    let vRes = new Vector3Builder(x, y, z);
    return vRes;
  }
};

// scripts/main.ts
var BUBBLE_BLACK_TAG = "ade:bubble_solution_black";
var BUBBLE_BLUE_TAG = "ade:bubble_solution_blue";
var BUBBLE_BROWN_TAG = "ade:bubble_solution_brown";
var BUBBLE_CYAN_TAG = "ade:bubble_solution_cyan";
var BUBBLE_GRAY_TAG = "ade:bubble_solution_gray";
var BUBBLE_GREEN_TAG = "ade:bubble_solution_green";
var BUBBLE_LIGHT_BLUE_TAG = "ade:bubble_solution_light_blue";
var BUBBLE_LIGHT_GRAY_TAG = "ade:bubble_solution_light_gray";
var BUBBLE_LIME_TAG = "ade:bubble_solution_lime";
var BUBBLE_MAGENTA_TAG = "ade:bubble_solution_magenta";
var BUBBLE_ORANGE_TAG = "ade:bubble_solution_orange";
var BUBBLE_PINK_TAG = "ade:bubble_solution_pink";
var BUBBLE_PURPLE_TAG = "ade:bubble_solution_purple";
var BUBBLE_RED_TAG = "ade:bubble_solution_red";
var BUBBLE_WHITE_TAG = "ade:bubble_solution_white";
var BUBBLE_YELLOW_TAG = "ade:bubble_solution_yellow";
var GLOWING_TAG = "ade:glowing";
var TICKS_PER_SECOND = 20;
var MOVEMENT_SIM_TICK_FREQ = 2;
var START_TICK = 100;
var curTick = 0;
var bubblesAndTargets = {};
system.runInterval(() => {
  curTick++;
  if (curTick > START_TICK && curTick % MOVEMENT_SIM_TICK_FREQ == 0) {
    simulateHoverAroundTarget();
  }
});
world.afterEvents.entityLoad.subscribe((event) => {
  if (event.entity.typeId == "ade:bubble") {
    event.entity.setProperty("ade:state", 1);
  }
});
system.afterEvents.scriptEventReceive.subscribe((e) => {
  if (e.id == "ade:player_spawn_bubble") {
    let dim = world.getDimension("overworld");
    const query = {
      name: e.message
    };
    let p = dim.getPlayers(query)[0];
    if (p) {
      spawnAndInitBubble(p);
    }
  } else if (e.id == "ade:bubble_leashed") {
    let leashedBubble = e.sourceEntity;
    if (leashedBubble && bubblesAndTargets[leashedBubble.id])
      delete bubblesAndTargets[leashedBubble.id];
  }
});
world.beforeEvents.entityRemove.subscribe((data) => {
  let id = data.removedEntity.id;
  if (data.removedEntity.typeId == "ade:bubble") {
    delete bubblesAndTargets[id];
  } else {
    system.run(() => {
      Object.keys(bubblesAndTargets).forEach((key) => {
        if (bubblesAndTargets[key].target.id == id) {
          bubblesAndTargets[key].bubble.setProperty("ade:state", 1);
          delete bubblesAndTargets[key];
        }
      });
    });
  }
});
function spawnAndInitBubble(p) {
  let viewDir = p.getViewDirection();
  let pRotation = p.getRotation();
  let bubbleSize = p.getProperty("ade:blown_bubble_size");
  let yOffset = 1.50625 - (bubbleSize ? (bubbleSize - 1) * 0.1 : 0);
  let playerModelOffset = "~0.0 ~" + yOffset.toString() + " ~0.0";
  let offsetFromFace = "^0 ^0 ^1.0";
  let bubbleSpawnCoords = p.location;
  bubbleSpawnCoords = MathUtils.applyDirectionalOffsetToCoords(bubbleSpawnCoords, playerModelOffset, viewDir);
  bubbleSpawnCoords = MathUtils.applyDirectionalOffsetToCoords(bubbleSpawnCoords, offsetFromFace, viewDir);
  let b = world.getDimension("minecraft:overworld").spawnEntity("ade:bubble<ade:eSummoned_invisible>", bubbleSpawnCoords);
  b.setRotation(pRotation);
  b.setProperty("ade:x_tilt", pRotation.x);
  let equipment = p.getComponent("equippable");
  let lEquip = equipment.getEquipment(EquipmentSlot.Offhand);
  if (lEquip) {
    let tags = lEquip.getTags();
    let skinSet = false;
    tags.forEach((tag) => {
      if (tag == BUBBLE_BLACK_TAG) {
        b.triggerEvent("ade:eBecome_black");
      } else if (tag == BUBBLE_BLUE_TAG) {
        b.triggerEvent("ade:eBecome_blue");
      } else if (tag == BUBBLE_BROWN_TAG) {
        b.triggerEvent("ade:eBecome_brown");
      } else if (tag == BUBBLE_CYAN_TAG) {
        b.triggerEvent("ade:eBecome_cyan");
      } else if (tag == BUBBLE_GRAY_TAG) {
        b.triggerEvent("ade:eBecome_gray");
      } else if (tag == BUBBLE_GREEN_TAG) {
        b.triggerEvent("ade:eBecome_green");
      } else if (tag == BUBBLE_LIGHT_BLUE_TAG) {
        b.triggerEvent("ade:eBecome_light_blue");
      } else if (tag == BUBBLE_LIGHT_GRAY_TAG) {
        b.triggerEvent("ade:eBecome_light_gray");
      } else if (tag == BUBBLE_LIME_TAG) {
        b.triggerEvent("ade:eBecome_lime");
      } else if (tag == BUBBLE_MAGENTA_TAG) {
        b.triggerEvent("ade:eBecome_magenta");
      } else if (tag == BUBBLE_ORANGE_TAG) {
        b.triggerEvent("ade:eBecome_orange");
      } else if (tag == BUBBLE_PINK_TAG) {
        b.triggerEvent("ade:eBecome_pink");
      } else if (tag == BUBBLE_PURPLE_TAG) {
        b.triggerEvent("ade:eBecome_purple");
      } else if (tag == BUBBLE_RED_TAG) {
        b.triggerEvent("ade:eBecome_red");
      } else if (tag == BUBBLE_WHITE_TAG) {
        b.triggerEvent("ade:eBecome_white");
      } else if (tag == BUBBLE_YELLOW_TAG) {
        b.triggerEvent("ade:eBecome_yellow");
      } else if (tag == GLOWING_TAG) {
        b.triggerEvent("ade:eSet_glowing_skin_type");
        skinSet = true;
      }
    });
    if (!skinSet) {
      b.triggerEvent("ade:eSet_default_skin_type");
    }
  }
  if (bubbleSize) {
    if (bubbleSize == 1)
      b.triggerEvent("ade:eBecome_size_1");
    else if (bubbleSize == 2)
      b.triggerEvent("ade:eBecome_size_2");
    else if (bubbleSize == 3)
      b.triggerEvent("ade:eBecome_size_3");
    else if (bubbleSize == 4)
      b.triggerEvent("ade:eBecome_size_4");
    else if (bubbleSize == 5)
      b.triggerEvent("ade:eBecome_size_5");
  }
  b.applyImpulse({ x: viewDir.x * 0.1, y: viewDir.y * 0.1, z: viewDir.z * 0.1 });
  let t = findTargetForBubble(b, p);
  if (t instanceof Entity) {
    b.setProperty("ade:state", 3);
    bubblesAndTargets[b.id] = { bubble: b, target: t, impulse: { x: 0, y: 0, z: 0 }, count: 10, maxCount: 10 };
  } else {
    b.setProperty("ade:state", 1);
  }
}
function findTargetForBubble(bubble, p) {
  let entities = p.getEntitiesFromViewDirection();
  if (entities) {
    let target = void 0;
    entities.forEach((e) => {
      if (e.entity.typeId != "ade:bubble")
        target = e.entity;
    });
    if (target) {
      return target;
    } else {
      return {};
    }
  } else {
    return {};
  }
}
function simulateHoverAroundTarget() {
  for (let key in bubblesAndTargets) {
    let curBubbleTargetPair = bubblesAndTargets[key];
    if (curBubbleTargetPair.bubble && curBubbleTargetPair.target) {
      generateImpulseAttempt3(curBubbleTargetPair);
    }
  }
}
function generateImpulseAttempt3(curBubbleTargetPair) {
  const TARGET_PERIMETER_X = 2 / MOVEMENT_SIM_TICK_FREQ;
  const TARGET_PERIMETER_Y = 2 / MOVEMENT_SIM_TICK_FREQ;
  const TARGET_PERIMETER_Z = 2 / MOVEMENT_SIM_TICK_FREQ;
  const TARGET_OUTER_PERIMETER_X = 4 / MOVEMENT_SIM_TICK_FREQ;
  const TARGET_OUTER_PERIMETER_Y = 4 / MOVEMENT_SIM_TICK_FREQ;
  const TARGET_OUTER_PERIMETER_Z = 4 / MOVEMENT_SIM_TICK_FREQ;
  const BUBBLE_SPEED = 1 / (TICKS_PER_SECOND / MOVEMENT_SIM_TICK_FREQ);
  let chasing_target_speed_multiplier = 1;
  let bubble = curBubbleTargetPair.bubble;
  let target = curBubbleTargetPair.target;
  if (bubble && target) {
    let bLoc = bubble.location;
    let tHeadLoc = void 0;
    try {
      tHeadLoc = target.getHeadLocation();
    } catch (e) {
    }
    if (tHeadLoc != void 0) {
      let impulse = new Vector3Builder(0, 0, 0);
      if (curBubbleTargetPair.count == curBubbleTargetPair.maxCount) {
        let count = 0;
        let maxCount = 0;
        let locDiff = Vector3Utils.subtract(tHeadLoc, bLoc);
        locDiff.x += (Math.random() - 0.5) / 2;
        locDiff.y += (Math.random() - 0.5) / 2;
        locDiff.z += (Math.random() - 0.5) / 2;
        impulse = Vector3Utils.normalize(locDiff);
        if (locDiff.x > TARGET_OUTER_PERIMETER_X || locDiff.y > TARGET_OUTER_PERIMETER_Y || locDiff.z > TARGET_OUTER_PERIMETER_Z || locDiff.x < -TARGET_OUTER_PERIMETER_X || locDiff.y < -TARGET_OUTER_PERIMETER_Y || locDiff.z < -TARGET_OUTER_PERIMETER_Z) {
          maxCount = 5;
          chasing_target_speed_multiplier = 2;
        } else {
          chasing_target_speed_multiplier = 1;
          maxCount = 1 + Math.floor(Math.random() * 50);
          let xDrift = Math.random() * 2 - 1;
          let yDrift = Math.random() * 2 - 1;
          let zDrift = Math.random() * 2 - 1;
          if (locDiff.x >= -TARGET_PERIMETER_X && locDiff.x <= TARGET_PERIMETER_X)
            impulse.x += xDrift;
          if (locDiff.y >= -TARGET_PERIMETER_Y && locDiff.y <= TARGET_PERIMETER_Y)
            impulse.y += yDrift;
          if (locDiff.z >= -TARGET_PERIMETER_Z && locDiff.z <= TARGET_PERIMETER_Z)
            impulse.z += zDrift;
          impulse = Vector3Utils.normalize(impulse);
          let xMovementCountMax = 0;
          if (impulse.x > 0 && bLoc.x + impulse.x * maxCount > tHeadLoc.x + TARGET_PERIMETER_X) {
            let distToPerimeter = tHeadLoc.x + TARGET_PERIMETER_X - bLoc.x;
            xMovementCountMax = Math.floor(distToPerimeter / BUBBLE_SPEED);
          } else if (impulse.x < 0 && bLoc.x + impulse.x * maxCount < tHeadLoc.x - TARGET_PERIMETER_X) {
            let distToPerimeter = bLoc.x - tHeadLoc.x + TARGET_PERIMETER_X;
            xMovementCountMax = Math.floor(distToPerimeter / BUBBLE_SPEED);
          }
          if (xMovementCountMax > 0) {
            maxCount = Math.min(maxCount, xMovementCountMax);
            maxCount = maxCount == 0 ? 1 : maxCount;
          }
          let yMovementCountMax = 0;
          if (impulse.y > 0 && bLoc.y + impulse.y * maxCount > tHeadLoc.y + TARGET_PERIMETER_Y) {
            let distToPerimeter = tHeadLoc.y + TARGET_PERIMETER_Y - bLoc.y;
            yMovementCountMax = Math.floor(distToPerimeter / BUBBLE_SPEED);
          } else if (impulse.y < 0 && bLoc.y + impulse.y * maxCount < tHeadLoc.y - TARGET_PERIMETER_Y) {
            let distToPerimeter = bLoc.y - tHeadLoc.y + TARGET_PERIMETER_Y;
            yMovementCountMax = Math.floor(distToPerimeter / BUBBLE_SPEED);
          }
          if (yMovementCountMax > 0) {
            maxCount = Math.min(maxCount, yMovementCountMax);
            maxCount = maxCount == 0 ? 1 : maxCount;
          }
          let zMovementCountMax = 0;
          if (impulse.z > 0 && bLoc.z + impulse.z * maxCount > tHeadLoc.z + TARGET_PERIMETER_Z) {
            let distToPerimeter = tHeadLoc.z + TARGET_PERIMETER_Z - bLoc.z;
            zMovementCountMax = Math.floor(distToPerimeter / BUBBLE_SPEED);
          } else if (impulse.z < 0 && bLoc.z + impulse.z * maxCount < tHeadLoc.z - TARGET_PERIMETER_Z) {
            let distToPerimeter = bLoc.z - tHeadLoc.z + TARGET_PERIMETER_Z;
            zMovementCountMax = Math.floor(distToPerimeter / BUBBLE_SPEED);
          }
          if (zMovementCountMax > 0) {
            maxCount = Math.min(maxCount, zMovementCountMax);
            maxCount = maxCount == 0 ? 1 : maxCount;
          }
        }
        impulse = Vector3Utils.scale(impulse, BUBBLE_SPEED * chasing_target_speed_multiplier);
        curBubbleTargetPair.impulse = impulse;
        curBubbleTargetPair.count = count;
        curBubbleTargetPair.maxCount = maxCount;
      } else {
        impulse = curBubbleTargetPair.impulse;
        curBubbleTargetPair.count++;
      }
      curBubbleTargetPair.bubble.clearVelocity();
      curBubbleTargetPair.bubble.applyImpulse(impulse);
    }
  }
}

//# sourceMappingURL=../debug/main.js.map
